/* Handles command line argument parsing and various option setup */

#ifndef _CMDLINE_H_
#define _CMDLINE_H_

/* Require some Personal Info to be Defined usually in config.h */
#ifndef PACKAGE
/* #error Please define macro "PACKAGE" */
#define PACKAGE "ftrace"
#endif

#ifndef VERSION
#error Please define macro "VERSION" */
//#define VERSION "0.X"
#endif


/* Function Declarations */
void usage();
void version();
void default_command_line(cmd_line* cl);
void pid_to_pathname(char* pid);
int parse_command_line(int argc, char **argv, cmd_line* cl);
int argv_to_execl(int argc, char **argv, int pos, char* fill[]);

/* Create a vararg function here for error reporting */
/*
fprintf(stderr,"%s: Cannot use both -e and -p options\n",PROGRAM);
usage();
exit(1);
*/

void usage()
{
	fprintf(stdout,"Usage: %s [options(s)] [command [arg ...]]\n"
	" Trace local functions, stack use and returns in [command] (./a.out)\n"
/*	" Options are [1=yes, 0=no] (default listed):\n" */
/*	"  -%c, --%-20s Trace command (a.out)\n" */
	"  -%c, --%-20s Attach to process pid rather than command (-%c0)\n"
	"  -%c, --%-20s List functions and return\n"
	"  -%c, --%-20s Start Trace at function, 0=none (main)\n"
	"  -%c, --%-20s End Trace at function, 0=none (main)\n"
	"  -%c, --%-20s Equivalent to -%c&-%c function (main)\n"
	"  -%c, --%-20s Break on selected function calls (1=yes)\n"
	"  -%c, --%-20s Show return statements (1=yes)\n"
	"  -%c, --%-20s Show verbose return information (0=no)\n"
	"  -%c, --%-20s Show num backtrace of stack at every func/ret (all)\n"
	"  -%c, --%-20s & List %%eip by: 0=no, 1=eip, 2=f_offset, (6=abs_f)\n"
	"  -%c, --%-20s List arguments to function calls (0=no)\n"
	"  -%c, --%-20s List return values from function calls (0=no)\n"
	"  -%c, --%-20s Toggle function arguments and return values (0=no)\n"  
	"  -%c, --%-20s Toggle almost everything: -%c%c%c%c%c%c%c (0=no)\n"
	"  -%c, --%-20s Redirect output to file (stderr)\n"
/*	"  -%c, --%-20s Filter selected functions\n" */
	"  -%c, --%-20s Show verbose information, use up to -vvv for more\n"
	"  -%c, --%-20s Display this program's version number\n"
	"  -%c, --%-20s Display the stream you are currently reading\n",PACKAGE,
/*	'e',"execute=[command]",	*/
	'p',"pid=[pid]",'S',
	'l',"list",
	'S',"start=[function]",
	'E',"end=[function]",
	'O',"only=[function]",'S','E',
	'c',"call=[0,1]",
	'r',"return=[0,1]",
	'R',"RETURN=[0,1]",
	's',"stack=[num|\'all\']",
	'i',"instruction=[0-3]",
	'a',"arguments=[0,1]",
	'z',"values=[0,1]",
	'Z',"VALUES=[0,1]",
	'x',"extra=[0,1]",'c','r','R','s','i','Z','v',
	'o',"output=[file]",
/*	'e',"expression=[expr]", */
	'v',"verbose",
	'V',"version",
	'h',"help"
	);
}

void version()
{fprintf(stdout,"%s -- version %s\n",PACKAGE,VERSION);}

void default_command_line(cmd_line* cl)
{
	cl->func_verbosity = 0;
	cl->begin_start = 1;
	cl->start_function=main_string;
	cl->end_ret = 1;
	cl->ret_function=main_string;
	cl->function_info = 0; /* 1=print functions, &2 arguments, &4 return */
	cl->normal_call = 1;
	cl->normal_ret = 1;
	cl->extra_ret = 0;
	cl->stack = -1;
	cl->show_eip = 6; /* 1=eip,2=func_offset,3=both,6=abs,0=no */
	cl->out_stream = stderr;
	cl->command = run_command;
	cl->child_pid = -1;
}

/* Parse command line arguments
   Return the value at argv in which we stop parsing
 */
int parse_command_line(int argc, char **argv, cmd_line* cl)
{
	int c;
	int option_index = 0;
	
	//Setup up default elements
	default_command_line(cl);
	
	static struct option long_options[] =
	{
/*	{"execute",required_argument,0,'e'},*/
	{"pid",required_argument,0,'p'},
	{"list",no_argument,0,'l'},
	{"start",required_argument,0,'S'},
	{"end",required_argument,0,'E'},
	{"only",required_argument,0,'O'},
	{"call",required_argument,0,'c'},
	{"return",required_argument,0,'r'},
	{"RETURN",required_argument,0,'R'},
	{"stack",required_argument,0,'s'},
	{"instructions",required_argument,0,'i'},
	{"arguments",required_argument,0,'a'},
	{"values",required_argument,0,'z'},
	{"VALUES",required_argument,0,'Z'},
	{"extra",required_argument,0,'x'},
	{"output",required_argument,0,'o'},
/*	{"expression",required_argument,0,'e'}, */
	{"verbose",no_argument,0,'v'},
	{"version",no_argument,0,'V'},
	{"help",no_argument,0,'h'},
	{0, 0, 0, 0}
	};
	/* man 3 getopt
	 If the first character of optstring is  ’+’  or  the
       environment  variable  POSIXLY_CORRECT  is  set, then option processing
       stops as soon as a non-option argument is encountered.
    */
	static char optstring[] = "+p:lS:E:O:c:r:R:s:i:a:z:Z:x:o:vVh";

	while(
	(c=getopt_long(argc,argv,optstring,long_options,&option_index))!=-1)
	{
		switch (c)
		{
			/* 'e' doesn't require an argument because
			 * we want to parse everything past it ourselves */
			/* both strace and ltrace use -e for expression, so we will do same
			case 'e':
				if (cl->child_pid!=-1)
				{
					fprintf(stderr,"ftrace: Cannot use both"
					 " -p and -e options\n");
					usage();
					exit(1);				
				}
				fprintf(stderr,"ftrace: optarg %s optind %i, argv[optind-1] %s\n",optarg,optind,argv[optind-1]);
				exit(0);
				//command = (char **) malloc((argc-optind+1) * sizeof(char *));
    	        //argv_to_execl(argc,argv,optind,command);	
				if (optind != argc)
					cl->command = argv[optind];
				if (argv[optind] == NULL)
				{
				    fprintf(stderr,"ftrace: please specify argument when using"
				     " the -e option.\n");
				}
				*/
				//return optind;
				//break;
			case 'p':
				cl->child_pid = atoi(optarg);
				if (cl->child_pid <= 0)
				{
					fprintf(stderr,"ftrace: "
					  "Invalid process id: %s\n",optarg);
					usage();
					exit(1);
				}
				cl->begin_start--;
				pid_to_pathname(optarg);
				break;
			case 'l':
				cl->function_info=1;
				break;
			case 'S':
				if (strncmp(optarg,"0",strlen(optarg))==0)
					cl->begin_start = 0;
				else
				{
				    cl->begin_start++;
					cl->start_function = optarg;
				}
				break;
			case 'E':
				if (strncmp(optarg,"0",strlen(optarg))==0)
					cl->end_ret = 0;
				else
				{
				    cl->end_ret++;
					cl->ret_function = optarg;
			    }
				break;
			case 'O':
				if (strncmp(optarg,"0",strlen(optarg))==0)
				{
					cl->begin_start = 0;
					cl->end_ret = 0;				
				}
				else
				{
				    cl->begin_start++;
				    cl->end_ret++;
					cl->start_function = optarg;
					cl->ret_function = optarg;
				}
				break;
			case 'c':
				cl->normal_call = atoi(optarg);
				break;
			case 'r':
				cl->normal_ret = atoi(optarg);
				break;
			case 'R':
				cl->extra_ret = atoi(optarg);
				break;
			case 's':
				if (strncmp(optarg,"all",strlen(optarg))==0)
					cl->stack = -1;
				else
					cl->stack = atoi(optarg);
				break;
			case 'i':
				cl->show_eip = atoi(optarg);
				break;
			case 'a':
			    if (atoi(optarg) == 0)
			        cl->function_info = cl->function_info & 5;
			    else
			        cl->function_info = cl->function_info | 2; 
				break;
			case 'z':
			    if (atoi(optarg) == 0)
			        cl->function_info = cl->function_info & 3;
			    else
			        cl->function_info = cl->function_info | 4; 
				break;
			case 'Z':
			    if (atoi(optarg) == 0)
			    {
			        cl->function_info = cl->function_info & 5;
			        cl->function_info = cl->function_info & 3;
			    }
			    else
			    {
			        cl->function_info = cl->function_info | 2;
			        cl->function_info = cl->function_info | 4; 
			    }
				break;			
			case 'x': /* 'c','r','R','s','P','a','z','v', */	
				cl->normal_call = atoi(optarg);
				cl->normal_ret = atoi(optarg);
				cl->extra_ret = atoi(optarg);
				
				if (atoi(optarg)>0)
				{
				    cl->begin_start = 0;
	                cl->end_ret = 0;
				    cl->show_eip = 3;
				    cl->stack = -1;
				    cl->function_info = cl->function_info | 2 | 4;
				    cl->func_verbosity+=3;
				}
				else
				{
				    cl->stack = 0;
				    cl->function_info = cl->function_info & 1;
				    cl->func_verbosity=0;
				   	cl->begin_start = 1;
                	cl->start_function=main_string;
                	cl->end_ret = 1;
				}
				break;
			case 'o':
			    if ((cl->out_stream = fopen(optarg,"w"))==NULL)
			    {
			         fprintf(stderr,"Error truncating file for writing %s\n",
			             optarg);
			         perror("fopen");
			         exit(errno);
			    }			
				break;
			case 'v':
				cl->func_verbosity++;
				break;
			case 'V':
				version();
				exit(0);
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case ':':
			case '?': //opterr, optopt
			default:
				usage();
				exit(1);
				break;		
		}	
	}
    /* if they specify a pid, just ignore any possible command (strace) */
	if (cl->child_pid==-1)
	{
		/* include below line if you want to be able to run without -e */
	   /* optind will be null if they run ftrace with no cmd line */
    	if (optind != argc) cl->command = argv[optind];

    	command = (char **) malloc((argc-optind+1) * sizeof(char *));
    	(*command) = cl->command; /* set command[0] manually if cmdline empty */
    	
        argv_to_execl(argc,argv,optind,command);
    }	
	
	return optind;
}

/* This function makes me paranoid, but is secure now */
/* how do we get this into cl->command */
void pid_to_pathname(char* pid)
{
    char *tmp_path;
    int string_size = strlen(PIDPRELUDE) + strlen(PIDLOCATION) + strlen(pid) +1;
    tmp_path = (char *) malloc(sizeof(char) * string_size);
    snprintf(tmp_path,string_size,"%s%i%s",PIDPRELUDE,cl->child_pid,PIDLOCATION);

    if (cl->func_verbosity>=1)
         fprintf(cl->out_stream,"Checking unresolved path: %s\n",tmp_path);
    
    cl->command = (char *) malloc(sizeof(char) * MAXPATHNAME);
    if ((string_size = readlink(tmp_path, cl->command, MAXPATHNAME-1)) == -1)
    {
        perror("readlink");
        exit(errno);
    }
    cl->command[string_size] = '\0';
    
    if (cl->func_verbosity>=1)
        fprintf(stderr,"Resolved Path is \'%s\'\n",cl->command);
}

int argv_to_execl(int argc, char **argv, int pos, char* fill[])
{
	int x=0;
	int i=pos;
	for(;i<argc;i++)
	{
	   fill[x++]=argv[i];
	   /* fprintf(cl->out_stream,"argv[%i]=%s\n",i,argv[i]); */
	}
	return x;
}

#endif /* _CMDLINE_H_ */
