/* Performs forking attaching and ptracing of another process */

#ifndef FTRACE_H_
#define FTRACE_H_

int start_trace();
void print_the_stack(int last);
void print_stack(int last);
int go_to_start(pid_t wait_val);
int report_functions();

int start_trace()
{
    /* Attach to a current pid rather than fork our own */
	if (cl->child_pid >= 0)
	{
		ATTACH(cl->child_pid);
 	    fprintf(cl->out_stream,"+++ %i attached to %i - interrupt to quit +++\n"
 	      ,getpid(),cl->child_pid);

		//kill(cl->child_pid,SIGSTOP);
	}

    /* Do we fork, or just start watching our attached pid */
	switch (cl->child_pid <= 0 ? cl->child_pid = fork() : cl->child_pid) 
	{
		case -1: /* error */
            	perror("fork");
            	break;
    	case 0: /* child process starts */   
            	//ptrace(PTRACE_TRACEME, 0, 0, 0); /* 7-12-06: Added */
    	        TRACE(0);	            	            	        
    	        //fprintf(stderr, "Executing Command: %s\n",cl->command);
    	        
    	        if (close_ioe()!=0)
                {
                    perror("closing stdin, stdout, stderr");
                    exit(errno);
                }                                
    		    if (execv(cl->command,command)==-1)
            		perror("execv");
                break;
        default: /* parent */
        	    if (cl->func_verbosity>=1)
        	         fprintf(cl->out_stream,"+++ %i beginning trace of %i +++\n"
        	           ,getpid(),cl->child_pid);
        	    /* if (cl->child_pid == 0) / Is this necessary?
        	        ATTACH(cl->child_pid); */
        		report_functions();
        		break;
    }
    
    DETACH(cl->child_pid);
    return 0;
}

int go_to_start(pid_t wait_val)
{
	ins = functions[get_start_function()].addr;
	
	if (cl->func_verbosity>=2)
	   fprintf(cl->out_stream,"Looking for start function: \'%s\'... ",
	       functions[get_start_function()].func_name);
	
	while (wait_val == WAIT_EXEC)
	{
		GET_REGS(cl->child_pid,&save_regs);
		
		if (ins == save_regs.eip)
		{
		    if (cl->func_verbosity>=2)
	           fprintf(cl->out_stream,"found!\n");
			return wait_val;
	    }
		
		STEPW(cl->child_pid,&wait_val);
    }
    
    if (cl->func_verbosity>=2)
       fprintf(cl->out_stream,"not found!\n");
	   
	return wait_val;
}

/* Returns 0 if everything went fine, 1 if we need to exit */
int check_asm() 
{
	int current_function = in_function(save_regs.eip);
	ins = PEEK(cl->child_pid,save_regs.eip);
	
	if (current_function == -1)
		return 0;	

	INS_IS_ASM(ins,RET)
	{
		if (cl->normal_ret > 0)
		{
			//if (cl->show_eip > 0)  print_addr(save_regs.eip);
			if (cl->stack != 0) print_stack(1);
			
			fprintf(cl->out_stream,"<- ");
			print_addr(save_regs.eip);
    		//fprintf(cl->out_stream," <-%s",functions[current_function].func_name);
    		  
    		if ((cl->function_info & 4) == 4)
            {
                fprintf(cl->out_stream,"(");
                print_value(save_regs.eax,MAXARRAY);
                fprintf(cl->out_stream,")");      
            }
            fprintf(cl->out_stream,"\n");
		}
		/*
		if (cl->extra_ret > 0)
		{
            //This should actually calculate the current address and report it
		    
			//if (cl->show_eip > 0) print_addr(PEEK(cl->child_pid,save_regs.esp));
			print_stack(1);
			//
			if ((cl->function_info & 4) == 4)
            {
                fprintf(cl->out_stream,"(");
                print_value(save_regs.eax,16);
                fprintf(cl->out_stream,")");      
            }
			fprintf(cl->out_stream,"\n");
			
		}
		*/
		if (cl->end_ret>0)
		{
			if (get_end_function() == current_function)
			{
				//if (cl->show_eip > 0)  print_addr(save_regs.eip);
				if (cl->func_verbosity>=2)
            	   fprintf(cl->out_stream,"Found end function: \'%s\'!\n",
            	       functions[get_end_function()].func_name);
            	
            	/* Print out the return value of the exiting function always */       
            	//if ((cl->function_info & 4) == 4)
                //{
                    fprintf(cl->out_stream,"+++ %i exited (status ",getpid());
                    print_value(save_regs.eax,16);
                    fprintf(cl->out_stream,") +++ \n");      
                //}

				return 1;
			}
		}
	
	}
	INS_IS_ASM(ins,CALL)
	{
		//Figure out where the call is going to
		if (in_function(save_regs.eip) != -1)
		{
			if (cl->func_verbosity>=3)
			{
			    /* ****************************************************************
			    if (cl->show_eip > 0)  print_addr(save_regs.eip);
			    
				long test;
				//Current IP + sizeof call + offset addr <= This isn't quite correct
				test = save_regs.eip +5+ ((0xFF00 & ins) >> 8);
				print_addr(save_regs.eip);
				fprintf(cl->out_stream,"C->");
				print_addr(test);
				fprintf(cl->out_stream,"\n");
				//ins = PEEK(cl->child_pid,save_regs.eip+4);
				*/
			}
		}
	}
	return 0;
}

int report_functions()
{
	pid_t wait_val;
	int f; int saw_ret = 0;
	int got_exit = 0;
	waitpid(cl->child_pid, &wait_val, WUNTRACED);
	
	//Grab the starting registers
	GET_REGS(cl->child_pid,&start_regs);
	
	/* Go from SIGSTOP to SIGTRAP, Complicated, but take 1|2 initial steps */
	/* This is to deal with some problems with the signal_handler */
	/* we can fix that by recognizing the difference between -p and -e */
	do{STEPW(cl->child_pid,&wait_val);}
	    while(wait_val == STOP_EXEC);
		
	/* Go to start function if specified */
	if (cl->begin_start>0)
		wait_val=go_to_start(wait_val);
	
	/*
    while (wait_val == WAIT_EXEC)
    {
        GET_REGS(cl->child_pid,&save_regs);
        
        if (AT_FUNCTION(f,save_regs.eip))        
        {
            fprintf(cl->out_stream,"%s <-",functions[f].func_name); 
            print_the_stack(0);                           
        }
        
        STEPW(cl->child_pid,&wait_val);        
    }
    if (got_exit==0)
    {
        GET_REGS(cl->child_pid,&save_regs); 
        //if (cl->show_eip > 0)  print_addr(save_regs.eip);
        check_wait_status(wait_val,cl->out_stream);fprintf(cl->out_stream,"\n");	
    }
    return 0;
    */
  
    
    ///////////////////////////////////////////////////////////////
	
	/* Wierd, this segfaults some code while readelf doesn't... */	
	while (wait_val == WAIT_EXEC)
    {
        GET_REGS(cl->child_pid,&save_regs);
        
        if (saw_ret != 0)
        {
            if ((f=in_function(save_regs.eip))!=-1)
            {
                if (cl->stack != 0) print_stack(0);
                print_addr(save_regs.eip);
                fprintf(cl->out_stream,"\n");
            }
            saw_ret = 0;  
        }
            	
    	// Check for particular instructions, and see if we exit
    	if ((got_exit = check_asm()) != 0)
    	   break; /* we hit end function */
    	
    	// Print when we land on a function
    	f = address_to_function(save_regs.eip);
    	if ((f != -1) && (cl->normal_call==1))
    	{       	
    		//if (cl->show_eip > 0) print_addr(save_regs.eip);	
		    if (cl->stack != 0) print_stack(0);
    		//fprintf(cl->out_stream,"%s",functions[f].func_name);
    		print_addr(save_regs.eip);
    		
    		// Print out possible arguments
    		if ((cl->function_info & 2) == 2)
    		  print_args();
    		
            /* don't print yet if we print asm instruction */
    		fprintf(cl->out_stream,"\n");
        }
        /* print out the assembly instruction here */
        
        /* Handle if (cl->extra_ret > 0) here */
        INS_IS_ASM(ins,RET)
        {
            if (cl->extra_ret > 0)
                saw_ret = 1;
        }
                
        STEPW(cl->child_pid,&wait_val);
    }
    
    /* Report the status of the child when we are done tracing */
    /* Only if check_asm does not return 1 (exit_function */ 
    if (got_exit==0)
    {
        GET_REGS(cl->child_pid,&save_regs); 
        //if (cl->show_eip > 0)  print_addr(save_regs.eip);
        check_wait_status(wait_val,cl->out_stream);fprintf(cl->out_stream,"\n");	
    }

    return 0;
}

void print_the_stack(int last)
{
	int n = 1;
	long f; int i;
	long ebp = save_regs.ebp;
	f = PEEK(cl->child_pid, ebp+4);
	
	//fprintf(cl->out_stream,"Traveling back down the stack:");
	
	while(ebp != 0)
	{			
		ebp = PEEK(cl->child_pid, ebp);	
		f = PEEK(cl->child_pid, ebp+4);	
		if ( f == 0xffffffff)
		    break;
		print_addr(f);
		fprintf(cl->out_stream,"<- ");
		n++;
	}
	fprintf(cl->out_stream,"\n");	
	//unsigned long* stack_address = (unsigned long *) malloc(n * sizeof(unsigned long));

}

/* Modify to find functions that we may not know about */
void print_stack(int last)
{
	int n = 1;
	long f; int i; //int j;
	long ebp = save_regs.ebp;
	f = PEEK(cl->child_pid, ebp+4);
	
	while(ebp != 0)
	{
	    //if (in_function(f) == -1) break;
	    //if (!is_func_space(f)) break;
	    //if (f==0) break;
	    
	    /* bug that somehow worked anyways
		if (address_to_function(f) == -1) break; */
					
		ebp = PEEK(cl->child_pid, ebp);	
		f = PEEK(cl->child_pid, ebp+4);
		//if (f != 0xffffffff)	
    	n++;
	}
	/* why not save this as addresses, so we can customize it more */
	unsigned long* stack_address = (unsigned long *) malloc(n * sizeof(unsigned long));
	   
	//char **names = (char **)malloc(n * sizeof(char **));
	//fprintf(cl->out_stream,"{%i}",n+1);
	
	n = 0;
	ebp = save_regs.ebp;
	
	/* If we are on a function, esp contains the previous function */
	//if (address_to_function(save_regs.eip) != -1)
	//{
		f = PEEK(cl->child_pid, save_regs.esp);
		//f = in_function(f);
		//if (f != -1)
		//if (in_function(f)!=-1)
		stack_address[n++] = f;
			//names[n++] = functions[f].func_name;
		//else names[n++] = unknown_function; /* unknown function */
			
	//}
	
	while(ebp != 0)
	{
		f = PEEK(cl->child_pid, ebp+4);
		ebp = PEEK(cl->child_pid, ebp);	
		
		//f = in_function(f);
		//if (f != -1)
		//if (in_function(f)!=-1)
		//if (f != 0xffffffff)
    	stack_address[n++] = f;
			//names[n++] = functions[f].func_name;
    	//else names[n++] = unknown_function; /* unknown function */
	}
	
	/* Used to guarentee that at least one stack element is printed
	 * even if cl->stack is 0, so we can use it for extra_ret */
	f = cl->stack!=0 ? cl->stack : 1;
	for(i=n-1; i >= 0 ; i--)
	{
		if (f > i || f == -1)
		{
		  
		    if (in_function(stack_address[i])!=-1)
		    {
		      print_addr(stack_address[i]);
			  if (i != 0 || last==0)
		          fprintf(cl->out_stream,"-> ");
		    }
		    else if (f != 0xffffffff)//unknown function
		    {
		      fprintf(cl->out_stream,"[0x%08x]",f);
      		  if (i != 0 || last==0)
		          fprintf(cl->out_stream,"-> ");
		    }
			/*	    

		    j = in_function(stack_address[i]);
		    //fprintf(cl->out_stream,"%i ",j);
			//if (j!=-1)//fprintf(cl->out_stream,"%s ",names[i]);

			if (j!=-1)
			{
			    fprintf(cl->out_stream,"%s ",functions[j].func_name);
			    
			    if (i != 0 || last==0)
				    fprintf(cl->out_stream,"->");
		    }
		    else // if you want to print out unknown addresses too
		    {
		        fprintf(cl->out_stream,"0x%08lx",stack_address[i]);
		        if (i != 0 || last==0)
				    fprintf(cl->out_stream,"->");
		    }
		    */
		    
		}
	}
	
	//free(names);
	free(stack_address);
}

#endif /* FTRACE_H_ */
