/* Functions to manipulate and work with elf binaries */

/* why this can't be named elf.h 
Error line:
 gcc -DHAVE_CONFIG_H -I. -I. -I.
 -I include directory 
 -I. include local directory
 Which means that elf.h in the local directory is getting included first
 Let this be a lesson for having the same name as a Clib include
*/

#ifndef MYELF_H_
#define MYELF_H_

/* Function Declarations */
Elf32_Half ehalf_swap(Elf32_Half half);
Elf32_Word eword_swap(Elf32_Word word);
int in_function(unsigned long address);
int address_to_function(unsigned long address);
int print_addr(unsigned long address);
int name_to_num(char *name);
int get_start_function();
int get_end_function();
void print_functions();
int function_parse(Elf32_Sym *symbol);
void get_functions(Elf32_Ehdr *header, char *image);
static void load_names(Elf32_Ehdr *header, char *image);
void check_ELF();
void close_elf();

/* Referenced from ELF Dump Written by Owen Klan */
/* Function to swap the byte-order for Elf32_Half items (ushorts) */
Elf32_Half ehalf_swap(Elf32_Half half)
{
    unsigned char ha[2];
#ifdef BIG_ENDIAN
    if (byte_order != ELFDATA2MSB)
      return half;
#else
    if (byte_order != ELFDATA2LSB)
      return half;
#endif
    
    ha[0] = (half & 0xff00) >> 8;
    ha[1] = (half & 0x00ff);
    return *(Elf32_Half *)ha;
}

/* Referenced from ELF Dump Written by Owen Klan */
/* Function to swap the byte-order for Elf32_Word items (ints) */
Elf32_Word eword_swap(Elf32_Word word)
{
    unsigned char wa[4];
    
#ifdef BIG_ENDIAN
    if (byte_order != ELFDATA2MSB)
      return word;
#else
    if (byte_order != ELFDATA2LSB)
      return word;
#endif
    
    wa[0] = (word & 0xff000000) >> 24;
    wa[1] = (word & 0x00ff0000) >> 16;
    wa[2] = (word & 0x0000ff00) >> 8;
    wa[3] = (word & 0x000000ff);
    
    return *(Elf32_Word *)wa;
}

/* few modifications, size <= replaced by < */
int in_function(unsigned long address)
{
	int i = -1;
	for(i=0; i < num_func; i++)
	{
	    /* direct match, we can return with these tests function */
        if (address == functions[i].addr)
           return i;
           
	    if (
	           (address >= functions[i].addr) && 
	           (address < (functions[i].size+functions[i].addr))
	       )
	       return i;
	}
	
	return -1;
}

int address_to_function(unsigned long address)
{
	int i;
	for(i=0; i < num_func; i++)
	{if (address==functions[i].addr)return i;}
	return -1;

}

int print_addr(unsigned long address)
{
	int i = in_function(address);
	int diff = -1;
	if (cl->show_eip == 0) /* do nothing */
	   return diff;
	if ((cl->show_eip & 1) == 1 || i == -1) /* print address */
	   fprintf(cl->out_stream,"0x%08lx",address); 
	if ((cl->show_eip & 3) == 3 || i == -1) /* both addr and function */
	   fprintf(cl->out_stream,":");
	if ((cl->show_eip & 2) == 2 && i != -1) /* print function if we can */
	{
	   fprintf(cl->out_stream,"%s",functions[i].func_name);
       diff = address - functions[i].addr;
    		
       /* Avoid <main+0> and if they want only absolute */
       if (diff != 0 && ((cl->show_eip & 6) != 6)) 
         fprintf(cl->out_stream,"+%d",diff);
	}
	
    return diff;
}

int name_to_num(char *name)
{
	int i;
	for(i=0; i < num_func; i++)
	{
	    if (strcmp(functions[i].func_name,name)==0)
	       return i;	
	}
	return -1;
}

int get_start_function()
{return name_to_num(cl->start_function);}

int get_end_function()
{return name_to_num(cl->ret_function);}

void print_functions()
{
	int i;
	fprintf(cl->out_stream, "\nFunction Information\n");
    fprintf(cl->out_stream, "%-3s  %-40s   %-10s %-6s %-8s\n",
        "Num", "Function", "Address", "Size", "Sec.Link");
 
	for(i=0; i < num_func; i++)
	{
		fprintf(cl->out_stream, "%3d  %-40s   0x%-8lx  %-6li  %-8x\n", i,
			functions[i].func_name, functions[i].addr,
			functions[i].size, functions[i].info);
	
	}
}

/* This function is largely not needed due to cmd_line -S/-E/-O switches */
/* when -e --expression is used, this will probably be re-enabled */
int function_parse(Elf32_Sym *symbol)
{
	//char* func_name = sym_name_table + eword_swap(symbol->st_name);
	
	//Avoid Dynamic Functions
	if (ehalf_swap(symbol->st_shndx)==0) return 0;
	
	//Avoid Empty Functions -> These will be sized at runtime
	//if (ehalf_swap(symbol->st_size)==0) return 0;

	//Avoid "Common" functions

	//if (func_name[0]=='_' && cl->func_verbosity<1){return 0;}
	//if (func_name[1]=='_' && cl->func_verbosity<2){return 0;}
	
	return 1;
}

void get_functions(Elf32_Ehdr *header, char *image)
{
    Elf32_Shdr *sectab;                /* section table offset */
    Elf32_Half sectab_size;            /* section table size */
    int num_syms;		       /* number of symbols */
    Elf32_Shdr *ptr;                   /* segment table buffer */
    Elf32_Shdr *sym_tab;	       /* current symbol */
    Elf32_Sym *symbol;
    int i;
    
    /* find the symbol table section */
    sectab = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    sectab_size = ehalf_swap(header->e_shnum) * ehalf_swap(header->e_shentsize);
        
    ptr = (Elf32_Shdr *)sectab;
    while (eword_swap(ptr->sh_type) != SHT_SYMTAB){ptr++;}

    num_syms = eword_swap(ptr->sh_size) / eword_swap(ptr->sh_entsize);
    sym_tab = (Elf32_Shdr *)(image + eword_swap(ptr->sh_offset));
    symbol = (void *)sym_tab;
    
    functions = (Elf_func*) malloc(num_syms * sizeof(Elf_func));
    
    /* now process all of the symbols, saving function items only */
    for (i=0; i < num_syms; i++) 
    {
	if (ELF32_ST_TYPE(symbol->st_info) == STT_FUNC) 
	{
	    char* func_name = sym_name_table + eword_swap(symbol->st_name);
	    
	    /* Pass function through regular expressions */
	    if (function_parse(symbol))
	    {
	        long addr = eword_swap(symbol->st_value);
	        long size = eword_swap(symbol->st_size);
	    	functions[num_func].func_name = func_name;
	    	functions[num_func].addr = addr;
	    	functions[num_func].size = size;
	    	functions[num_func].info = eword_swap(symbol->st_info);
	    	num_func++;
	    	
	    	if (addr < first_func_addr)
	    	   first_func_addr = addr;

   	    	/* Find the max_possible addressable function */
   	    	if ((addr + size) > last_func_addr)
	    	   last_func_addr = (addr + size);	    	
	    }	    
	}
	symbol++;
    }
    
    /*
    fprintf(stderr,"Function Addresses are between 0x%0x-0x%0x\n",
                   first_func_addr,last_func_addr);
    */
    
    /* Print out function info and return */
    if (cl->function_info==1)
    {
    	print_functions();
    	exit(1);
    }
    
    /* Check that a start and end function exist if chosen */
    if (get_start_function(functions, num_func)==-1 && cl->begin_start>0)
    {
    	fprintf(stderr,
    	  "Error: No start function named: \'%s\'\n",cl->start_function);
    	print_functions();
    	exit(1);    
    }
    if (get_end_function()==-1 && cl->end_ret>0)
    {
        fprintf(stderr,
          "Error: No end function named: \'%s\'\n",cl->ret_function);
        print_functions();
    	exit(1); 
    
    }

    return;
}

static void load_names(Elf32_Ehdr *header, char *image)
{
    Elf32_Shdr *ptr;
    int i;

    ptr = (Elf32_Shdr *)
      (image + eword_swap(header->e_shoff)) + ehalf_swap(header->e_shstrndx);
    
    name_table = (char *)(image + eword_swap(ptr->sh_offset));
    
    /* load the general string table */
    ptr = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    for (i=(int)ehalf_swap(header->e_shnum); i > 0 ; i--)
    {
	if (strncmp((name_table + eword_swap(ptr->sh_name)),".strtab", 7) == 0)
	{
	    sym_name_table = (char *)(image + eword_swap(ptr->sh_offset));
	} 
	ptr++;
    }
    
    if (sym_name_table == NULL)
    {
    	fprintf(stderr, 
    	  "\nNo symbol table defined: Striped Binary: Quitting.\n\n");
    	exit(1);
    }
}

void check_ELF()
{
    Elf32_Ehdr *elf_head;	       /* ELF header structure */

    /* check that the file exists */
    if ((elf_fp = open(cl->command, O_RDONLY)) < 0)
    {
	    fprintf(stderr, 
	     "Could not open %s. Perror: %s\n", cl->command, strerror(errno));
	    exit(errno);
    }
    
    /* use fstat() to get the file size */
    if (fstat(elf_fp, &elf_stat) < 0)
    {
	    fprintf(stderr, 
	     "Could not stat file pointer. Perror: %s\n",strerror(errno));
	    exit(errno);
    }
    
    /* attempt to mmap() the file to an address */
    if ((elf_image = 
    	mmap(0, elf_stat.st_size, PROT_READ, MAP_PRIVATE, elf_fp, 0)
    	) == MAP_FAILED) 
    {
	    fprintf(stderr, "Failure on mmap. Perror: %s\n", strerror(errno));
	    exit(errno);
    }
    
    /* check for magic header */  
    if (strncmp(elf_image, ELFMAG, 3) != 0)
    {
    	fprintf(stderr, "%s is not an ELF file.\n", cl->command);
	    exit(1);
    }
    
    elf_head = (Elf32_Ehdr *)elf_image;
    byte_order = elf_head->e_ident[EI_DATA];  
    
    /* get the string table */
    load_names(elf_head, elf_image);
    /* get the functions for our trace */
    get_functions(elf_head, elf_image);

}

void close_elf()
{
    /* Clean up and free our file info */
    if (munmap(elf_image, elf_stat.st_size) < 0)
    {
	    fprintf(stderr, "Failed unmapping ELF file: %s",strerror(errno));
	    exit(errno);
    }
    
    if (close(elf_fp)==-1)
    {
    	fprintf(stderr, "Error closing ELF file: %s",strerror(errno));
	    exit(errno); 
    }
}

#endif /* MYELF_H_ */
